/* Name: main.c
 * Project: AVR MIDI device on Low-Speed USB
 * Author: Yoshitaka Kuwata
 * Based on : Martin Homuth-Rosemann
 * Creation Date: 2008-03-11
 * Copyright: (c) 2008 by Martin Homuth-Rosemann.
 * License: GPL v2 or later.
 *
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbdrv.h"
#include "oddebug.h"


// This deviceDescrMIDI[] is based on 
// http://www.usb.org/developers/devclass_docs/midi10.pdf
// Appendix B. Example: Simple MIDI Adapter (Informative)

// B.1 Device Descriptor
static PROGMEM char deviceDescrMIDI[] = {	/* USB device descriptor */
	18,			/* sizeof(usbDescriptorDevice): length of descriptor in bytes */
	USBDESCR_DEVICE,	/* descriptor type */
	0x10, 0x01,		/* USB version supported */
	0,			/* device class: defined at interface level */
	0,			/* subclass */
	0,			/* protocol */
	8,			/* max packet size */
	USB_CFG_VENDOR_ID,	/* 2 bytes */
	USB_CFG_DEVICE_ID,	/* 2 bytes */
	USB_CFG_DEVICE_VERSION,	/* 2 bytes */
	1,			/* manufacturer string index */
	2,			/* product string index */
	0,			/* serial number string index */
	1,			/* number of configurations */
};

// B.2 Configuration Descriptor
static PROGMEM char configDescrMIDI[] = {	/* USB configuration descriptor */
	9,			/* sizeof(usbDescrConfig): length of descriptor in bytes */
	USBDESCR_CONFIG,	/* descriptor type */
	101, 0,			/* total length of data returned (including inlined descriptors) */
	2,			/* number of interfaces in this configuration */
	1,			/* index of this configuration */
	0,			/* configuration name string index */
#if USB_CFG_IS_SELF_POWERED
	USBATTR_SELFPOWER,	/* attributes */
#else
	USBATTR_BUSPOWER,	/* attributes */
#endif
	USB_CFG_MAX_BUS_POWER / 2,	/* max USB current in 2mA units */

// B.3 AudioControl Interface Descriptors
// The AudioControl interface describes the device structure (audio function topology) 
// and is used to manipulate the Audio Controls. This device has no audio function 
// incorporated. However, the AudioControl interface is mandatory and therefore both 
// the standard AC interface descriptor and the classspecific AC interface descriptor 
// must be present. The class-specific AC interface descriptor only contains the header 
// descriptor.

// B.3.1 Standard AC Interface Descriptor
// The AudioControl interface has no dedicated endpoints associated with it. It uses the 
// default pipe (endpoint 0) for all communication purposes. Class-specific AudioControl 
// Requests are sent using the default pipe. There is no Status Interrupt endpoint provided.
	/* descriptor follows inline: */
	9,			/* sizeof(usbDescrInterface): length of descriptor in bytes */
	USBDESCR_INTERFACE,	/* descriptor type */
	0,			/* index of this interface */
	0,			/* alternate setting for this interface */
	0,			/* endpoints excl 0: number of endpoint descriptors to follow */
	1,			/* */
	1,			/* */
	0,			/* */
	0,			/* string index for interface */

// B.3.2 Class-specific AC Interface Descriptor
// The Class-specific AC interface descriptor is always headed by a Header descriptor 
// that contains general information about the AudioControl interface. It contains all 
// the pointers needed to describe the Audio Interface Collection, associated with the 
// described audio function. Only the Header descriptor is present in this device 
// because it does not contain any audio functionality as such.
	/* descriptor follows inline: */
	9,			/* sizeof(usbDescrCDC_HeaderFn): length of descriptor in bytes */
	36,			/* descriptor type */
	1,			/* header functional descriptor */
	0x0, 0x01,		/* bcdADC */
	9, 0,			/* wTotalLength */
	1,			/* */
	1,			/* */

// B.4 MIDIStreaming Interface Descriptors

// B.4.1 Standard MS Interface Descriptor
	/* descriptor follows inline: */
	9,			/* length of descriptor in bytes */
	USBDESCR_INTERFACE,	/* descriptor type */
	1,			/* index of this interface */
	0,			/* alternate setting for this interface */
	2,			/* endpoints excl 0: number of endpoint descriptors to follow */
	1,			/* AUDIO */
	3,			/* MS */
	0,			/* unused */
	0,			/* string index for interface */

// B.4.2 Class-specific MS Interface Descriptor
	/* descriptor follows inline: */
	7,			/* length of descriptor in bytes */
	36,			/* descriptor type */
	1,			/* header functional descriptor */
	0x0, 0x01,		/* bcdADC */
	65, 0,			/* wTotalLength */

// B.4.3 MIDI IN Jack Descriptor
	/* descriptor follows inline: */
	6,			/* bLength */
	36,			/* descriptor type */
	2,			/* MIDI_IN_JACK desc subtype */
	1,			/* EMBEDDED bJackType */
	1,			/* bJackID */
	0,			/* iJack */

	/* descriptor follows inline: */
	6,			/* bLength */
	36,			/* descriptor type */
	2,			/* MIDI_IN_JACK desc subtype */
	2,			/* EXTERNAL bJackType */
	2,			/* bJackID */
	0,			/* iJack */

//B.4.4 MIDI OUT Jack Descriptor
	/* descriptor follows inline: */
	9,			/* length of descriptor in bytes */
	36,			/* descriptor type */
	3,			/* MIDI_OUT_JACK descriptor */
	1,			/* EMBEDDED bJackType */
	3,			/* bJackID */
	1,			/* No of input pins */
	2,			/* BaSourceID */
	1,			/* BaSourcePin */
	0,			/* iJack */

	/* descriptor follows inline: */
	9,			/* bLength of descriptor in bytes */
	36,			/* bDescriptorType */
	3,			/* MIDI_OUT_JACK bDescriptorSubtype */
	2,			/* EXTERNAL bJackType */
	4,			/* bJackID */
	1,			/* bNrInputPins */
	1,			/* baSourceID (0) */
	1,			/* baSourcePin (0) */
	0,			/* iJack */


// B.5 Bulk OUT Endpoint Descriptors

//B.5.1 Standard Bulk OUT Endpoint Descriptor
	/* descriptor follows inline: */
	9,			/* bLenght */
	USBDESCR_ENDPOINT,	/* bDescriptorType = endpoint */
	0x1,			/* bEndpointAddress OUT endpoint number 1 */
	3,			/* bmAttributes: 2:Bulk, 3:Interrupt endpoint */
	8, 0,			/* wMaxPacketSize */
	10,			/* bIntervall in ms */
	0,			/* bRefresh */
	0,			/* bSyncAddress */

// B.5.2 Class-specific MS Bulk OUT Endpoint Descriptor
	/* descriptor follows inline: */
	5,			/* bLength of descriptor in bytes */
	37,			/* bDescriptorType */
	1,			/* bDescriptorSubtype */
	1,			/* bNumEmbMIDIJack  */
	1,			/* baAssocJackID (0) */


//B.6 Bulk IN Endpoint Descriptors

//B.6.1 Standard Bulk IN Endpoint Descriptor
	/* descriptor follows inline: */
	9,			/* bLenght */
	USBDESCR_ENDPOINT,	/* bDescriptorType = endpoint */
	0x81,			/* bEndpointAddress IN endpoint number 1 */
	3,			/* bmAttributes: 2: Bulk, 3: Interrupt endpoint */
	8, 0,			/* wMaxPacketSize */
	10,			/* bIntervall in ms */
	0,			/* bRefresh */
	0,			/* bSyncAddress */

// B.6.2 Class-specific MS Bulk IN Endpoint Descriptor
	/* descriptor follows inline: */
	5,			/* bLength of descriptor in bytes */
	37,			/* bDescriptorType */
	1,			/* bDescriptorSubtype */
	1,			/* bNumEmbMIDIJack (0) */
	3,			/* baAssocJackID (0) */
};


uchar usbFunctionDescriptor(usbRequest_t * rq)
{
	if (rq->wValue.bytes[1] == USBDESCR_DEVICE) {
		usbMsgPtr = (uchar *) deviceDescrMIDI;
		return sizeof(deviceDescrMIDI);
	} else {		/* must be config descriptor */
		usbMsgPtr = (uchar *) configDescrMIDI;
		return sizeof(configDescrMIDI);
	}
}


static uchar sendEmptyFrame;


/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

uchar usbFunctionSetup(uchar data[8])
{
  usbRequest_t *rq = (void *) data;

  if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {	/* class request type */

    /*  Prepare bulk-in endpoint to respond to early termination   */
    if ((rq->bmRequestType & USBRQ_DIR_MASK) ==
	USBRQ_DIR_HOST_TO_DEVICE)
      sendEmptyFrame = 1;
  }
  
  return 0xff;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionRead                                                           */
/*---------------------------------------------------------------------------*/

uchar usbFunctionRead(uchar * data, uchar len)
{
	data[0] = 0;
	data[1] = 0;
	data[2] = 0;
	data[3] = 0;
	data[4] = 0;
	data[5] = 0;
	data[6] = 0;

	return 7;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionWrite                                                          */
/*---------------------------------------------------------------------------*/

uchar usbFunctionWrite(uchar * data, uchar len) {
	return 1;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionWriteOut                                                       */
/*                                                                           */
/* this Function is called if a MIDI Out message (from PC) arrives.          */
/*                                                                           */
/*---------------------------------------------------------------------------*/

void usbFunctionWriteOut(uchar * data, uchar len) {
}



/*---------------------------------------------------------------------------*/
/* hardwareInit                                                              */
/*---------------------------------------------------------------------------*/

static void hardwareInit(void) {
  uchar i, j;

  /* activate pull-ups except on USB lines */
  USB_CFG_IOPORT =
    (uchar) ~ ((1 << USB_CFG_DMINUS_BIT) |
	       (1 << USB_CFG_DPLUS_BIT));
  /* all pins input except USB (-> USB reset) */
#ifdef USB_CFG_PULLUP_IOPORT
  /* use usbDeviceConnect()/usbDeviceDisconnect() if available */
  USBDDR = 0;		/* we do RESET by deactivating pullup */
  usbDeviceDisconnect();
#else
  USBDDR = (1 << USB_CFG_DMINUS_BIT) | (1 << USB_CFG_DPLUS_BIT);
#endif

  j = 0;
  while (--j) {		/* USB Reset by device only required on Watchdog Reset */
    i = 0;
    while (--i);	/* delay >10ms for USB reset */
  }
#ifdef USB_CFG_PULLUP_IOPORT
  usbDeviceConnect();
#else
  USBDDR = 0;		/*  remove USB reset condition */
#endif

  /* ポートの初期化をする */
  DDRB &= ~0b11011110;		/* INPUT */
  DDRB |=  0b00100001;		/* OUTPUT */
  DDRC =   0b00001111;		/* OUTPUT */
  DDRD &= ~0b11111010;		/* INPUT */

  PORTB = 0b00010001;		/* pullup PB4 for test button key input */
  PORTC = 0b00001111;
  PORTD = 0b11111010;		/* pullup PD for button input */
}

#define setPortBit(port, bit, val) { \
    if (val == 0) { *(port) &= ~_BV(bit); } else	\
      {*(port) |= _BV(bit); }				\
}
#define getPortBit(port, bit) ((*(port) & _BV(bit)) >> (bit))

/* SWITCH */
#define gioGetSW(void) getPortBit(&PINB, 4)

#define gioSetLED(val) setPortBit(&PORTB, 5, (val))

/* ketEventのQueue */
struct keyEvent {
  uchar key;
  uchar onOff;
};

#define KEY_OFF 0
#define KEY_ON 1
#define MAX_KEY_EVENT 32
#define KEY_EVENT_QUEUE_MASK (MAX_KEY_EVENT-1)
struct keyEvent keyEventQueue[MAX_KEY_EVENT];
uchar urptr, uwptr, iwptr;

uchar keyEventsFree (struct keyEvent **ptr) {
  return (MAX_KEY_EVENT - uwptr - 1);
}

void keyEventAppend(uchar key, uchar onOff) {
  uchar iwnxt;
  iwnxt = (iwptr +1 ) & KEY_EVENT_QUEUE_MASK;
  if (iwnxt == urptr )
    return;			/* buffer overflow */
  keyEventQueue[iwptr].key = key;
  keyEventQueue[iwptr].onOff = onOff;
  iwptr = iwnxt;
}

static inline uchar keyEventAvailable(struct keyEvent **ptr) {
  *ptr = &keyEventQueue[urptr];
  if(iwptr >= urptr) {
    return iwptr - urptr;
  } else {  /* buffer end is between read and write pointer, return continuous range */
    return MAX_KEY_EVENT - urptr;
  }
}

/* The following function must be called after uartRxBytesAvailable() to
 * remove the bytes from the receiver buffer.
 */
static inline void  keyEventDidReadEvents(uchar numEvents) {
  urptr = (urptr + numEvents) & KEY_EVENT_QUEUE_MASK;
}

static inline void keyEventQueueClear() {
  urptr = uwptr = 0;
}

struct portInfo {
  volatile uchar *port;
  uchar bit;
};

#define N_SCAN_COL 5
static struct portInfo scanOut[] = {
  {&PORTB, 0}, {&PORTC, 0}, {&PORTC, 1}, {&PORTC, 2}, {&PORTC, 3}
};

static uchar keyTable[][6] = {
  {60, 61, 62, 63, 64, 65},
  {66, 67, 68, 69, 70, 71},
  {72, 73, 74, 75, 76, 77},
  {78, 79, 80, 81, 82, 83},
  {84, 85, 86, 87, 88, 89}
};

static uchar keyStatus [N_SCAN_COL]; /* 前のキーの値 */
static uchar keyCount [5][6]; /* 同じキーが続いた回数を数える */

/* チャッタリング防止用に読み込む回数 */
#define KEY_COUNT_TH 10

void initializeKeyStatus() {
  int i, j;
  for (i = 0; i < N_SCAN_COL; i++) {
    *(scanOut[i].port) |= _BV(scanOut[i].bit);	/* set HIGH */
    keyStatus[i] = 0b00111111;
    for (j = 0 ; j < 6 ; j++) 
      keyCount[i][j] = 0;
  }
}

static uchar scanKeyPressed(void) {
  uchar i, j, x;
  uchar bit;
 for (i = 0 ; i < N_SCAN_COL ; i++) {
    /* read key status*/
    *(scanOut[i].port) &= ~_BV(scanOut[i].bit);	/* set LOW */
     for (j = 0 ; j < 100 ; j++) {
       x = x;
    }
    x = PIND;					/* need to wait */
    x = PIND;					/* need to wait */
    x = ((x >>2) & 0b00111110) | ((x >>1) & 0b00000001); /* bit寄せ */
    *(scanOut[i].port) |= _BV(scanOut[i].bit);	/* set HIGH */

    for (j = 0 ; j < 6 ; j++) {
      bit = _BV(j);
      if (((bit & keyStatus[i]) != (bit & x)) && ((keyCount[i][j]++) > KEY_COUNT_TH)) {
	if (((bit & keyStatus[i]) != 0) && ((bit & x) == 0)) {
	  /* key pressed */
	  keyEventAppend(keyTable[i][j], 1);
	  keyStatus[i] &= ~bit;		/* save new status */
	} else if (((bit & keyStatus[i]) == 0) && ((bit & x) != 0)) {
	  /* key released */
	  keyEventAppend(keyTable[i][j], 0);
	  keyStatus[i] |= bit;		/* save new status */
	}
	keyCount[i][j] = 0;
	gioSetLED(1);
      } else {
	gioSetLED(0);
      }
    }
  }
  return 0;
}



int main(void)
{
  uchar midiMsg[8];
  uchar i, iii;

  wdt_enable(WDTO_1S);
  hardwareInit();
  odDebugInit();
  usbInit();

  sendEmptyFrame = 0;
  initializeKeyStatus();
  keyEventQueueClear();
  sei();

  uchar prekey = 1;
  uchar keycount = 0;
  for (;;) {		/* main event loop */
    //    PORTC ^= 0x40;	// DEBUG LED
    wdt_reset();
    usbPoll();

    /* キーボードをスキャンする */
    scanKeyPressed();

/*     /\* test *\/ */
/*     if ((prekey == 1) && (gioGetSW() == 0)) { */
/*       if ((keycount++) > 10) { */
/* 	keyEventAppend(60, 1); */
/* 	prekey = 0; */
/* 	gioSetLED(1); */
/*       } */
/*     } else if ((prekey == 0) && (gioGetSW() != 0)) { */
/*       if ((keycount++) > 10) { */
/* 	keyEventAppend(60, 0); */
/* 	prekey = 1; */
/* 	gioSetLED(1); */
/*       } */
/*     } else { */
/*       keycount = 0; */
/*       gioSetLED(0); */
/*     } */

    /* USBに送信する */
    if (usbInterruptIsReady()) {
      uchar eventsRead;
      struct keyEvent *data;
      
      eventsRead = keyEventAvailable(&data);
      if(eventsRead > 0 || sendEmptyFrame){
	if(eventsRead >= 2) {	/* up to 2 events can be sent */
	  eventsRead = 2;
	  /* send an empty block after last data block to indicate transfer end */
	  sendEmptyFrame  = 1;
	} else {
	  sendEmptyFrame  = 0;
	}
	iii = 0;
	for (i = 0; i < eventsRead ; i++) {
	  if ((data ->onOff) == KEY_ON) { /* NOTE ON */
	    midiMsg[iii++] = 0x09;
	    midiMsg[iii++] = 0x90;
	    midiMsg[iii++] = data->key;
	    midiMsg[iii++] = 0x7f;
	  } else {		/* NOTE OFF */
	    midiMsg[iii++] = 0x08;
	    midiMsg[iii++] = 0x80;
	    midiMsg[iii++] = data->key;
	    midiMsg[iii++] = 0x00;
	  }
	}
	usbSetInterrupt(midiMsg, eventsRead * 4);
	keyEventDidReadEvents(eventsRead);
      }
    }		// usbInterruptIsReady()
  }			// main event loop
  return 0;
}
